/* My solution to the "Lisp as an Alternative to Java" challenge.
 * See http://www.flownet.com/ron/papers/lisp-java/
 * 
 * Tomer Gabel, June 10th 2007 
 * 
 * Total coding time, including optimization, cleanup and comments: 2 hours 20 minutes
 * 
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;

namespace TomerGabel.Challenge
{
    /// <summary>
    /// Provides the actual phone number encoder implementation
    /// </summary>
    class Encoder
    {
        
        /// <summary>
        /// Maps characters (a-z) to their respective values, according to the provided digits. Should be indexed according
        /// to offset from ASCII 'a' (i.e. 'a' is 0, 'b' is 1 etc.)
        /// </summary>
        //                                      a, b, c, d, e, f, g, h, i. j, k, l, m, n, o, p, q, r, s, t, u, v, w, x, y, z
        private byte[] m_charmap = new byte[] { 5, 7, 6, 3, 0, 4, 9, 9, 6, 1, 7, 8, 5, 1, 8, 8, 1, 2, 3, 4, 7, 6, 2, 2, 3, 9 };

        /// <summary>
        /// Castrated decimal (10-branch) tree implementation. Each node has 10 optional branches, as well as a list of leaves.
        /// </summary>
        private class DecimalTree
        {
            private DecimalTree[] m_branches = new DecimalTree[ 10 ];

            /// <summary>
            /// List of leaves
            /// </summary>
            public List<string> Leaves = new List<string>();

            /// <summary>
            /// Returns the specified branch. If the branch does not yet exist, it is allocated.
            /// </summary>
            /// <param name="index">Index of the accessed branch.</param>
            /// <returns>The specified branch.</returns>
            public DecimalTree Branch( int index )
            {
                return m_branches[ index ] ?? ( m_branches[ index ] = new DecimalTree() );
            }
        }

        /// <summary>
        /// Receives the string dictionary.
        /// </summary>
        /// <remarks>To find a valid dictionary word, traverse the tree according to the character map (<see cref="m_charmap"/>);
        /// the leaves represent the words that match the phone number up until this point in the traversal.</remarks>
        private DecimalTree m_dictionary = new DecimalTree();

        /// <summary>
        /// Creates and initializes an encoder instance.
        /// </summary>
        /// <param name="reader">The input file stream from which to derive the dictionary.</param>
        public Encoder( StreamReader reader )
        {
            // Parse the input dictionary
            while ( !reader.EndOfStream )
            {
                string entry = reader.ReadLine();

                CharEnumerator en = entry.ToLower().GetEnumerator();
                DecimalTree node = m_dictionary;
                while ( en.MoveNext() )
                    if ( en.Current >= 'a' && en.Current <= 'z' )
                        node = node.Branch( m_charmap[ en.Current - 'a' ] );

                node.Leaves.Add( entry );
            }
            
        }

        /// <summary>
        /// Converts a string phone number representation into the underlying list of values.
        /// </summary>
        /// <param name="input">Input phone number in string form.</param>
        /// <returns>Normalized phone number, without dashes, slashes or any of the other pesky characters.</returns>
        private List<byte> normalizeInput( string input )
        {
            List<byte> builder = new List<byte>( input.Length );
            foreach ( char c in input )
                if ( c >= '0' && c <= '9' )
                    builder.Add( (byte) ( c - '0' ) );
            return builder;
        }

        /// <summary>
        /// Returns all encoded forms of the specified phone number.
        /// </summary>
        /// <param name="input">Phone number to be encoded.</param>
        /// <returns>An array of valid, dictionary-encoded string for the specified phone number.</returns>
        public string[] Encode( string input )
        {
            return encode( normalizeInput( input ), true ).ToArray();
        }

        /// <summary>
        /// Returns all encoded forms of the specified phone number (or partial string); optionally processes single digits as well.
        /// </summary>
        /// <param name="input">Partial phone number to be encoded.</param>
        /// <param name="processDigit"><c>True</c> if single-digits are allowed at this point, <c>false</c> otherwise.</param>
        /// <returns></returns>
        public List<string> encode( List<byte> input, bool processDigit )
        {
            List<string> results = new List<string>();

            // Empty inputs return an empty list
            if ( input.Count == 0 )
                return results;

            // Traverse the dictionary tree in the hope of finding suitable dictionary words
            bool foundAny = false;
            List<byte>.Enumerator en = input.GetEnumerator();
            DecimalTree node = m_dictionary;
            int nesting = 0;
            while ( en.MoveNext() )
            {
                // Move to the next relevant branch in the tree
                node = node.Branch( en.Current );
                ++nesting;

                // If there are any valid leaves it means we have suitable dictionary words
                if ( node.Leaves.Count > 0 )
                {
                    foundAny = true;

                    // If the nesting level (word length) is the same as the input length, it means out work is done. Otherwise
                    // we'll just split the relevant part of the input and recurse over the possibilities for that.
                    if ( nesting == input.Count )
                        results.AddRange( node.Leaves );
                    else
                        foreach ( string leaf in node.Leaves )
                        foreach ( string subtext in encode( input.GetRange( nesting, input.Count - nesting ), true ) )
                            results.Add( leaf + " " + subtext );
                }
            }

            // If no suitable words are found and digits are allowed, add the digit to the encoded result and keep processing.
            if ( processDigit && !foundAny )
            {
                byte digit = input[ 0 ];
                if ( input.Count == 1 )
                    results.Add( digit.ToString() );
                else
                    foreach ( string s in encode( input.GetRange( 1, input.Count - 1 ), false ) )
                        results.Add( digit + " " + s );
            }

            // All done, time to return our result set.
            return results;
        }
    }

    /// <summary>
    /// Entry-point and running script for the encoder
    /// </summary>
    class Driver
    {
        static void Main( string[] args )
        {
            // Check for valid parameters
            if ( args.Length < 2 )
            {
                Console.WriteLine( "Usage: " + Assembly.GetEntryAssembly().GetName().Name + " <dictionary_file> <input_file>" );
                return;
            }

            // Instantiate encoder from specified dictionary file
            Encoder encoder;
            using ( StreamReader sr = File.OpenText( args[ 0 ] ) )
                encoder = new Encoder( sr );

            // Iterate input file and dump encoded strings
            using ( StreamReader sr = File.OpenText( args[ 1 ] ) )
                while ( !sr.EndOfStream )
                {
                    string input = sr.ReadLine();

                    foreach ( string encodedEntry in encoder.Encode( input ) )
                        Console.WriteLine( input + ": " + encodedEntry );
                }
        }
    }
}
